// hersrch.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <signal.h>
#include "BasicComponent.h"
#include "CompoundComponent.h"
#include "Search.h"
#include "Writer.h"
#include "Identifier.h"
#include <time.h>

static Search * s_search = NULL;

void handleSigInt(int sig) 
{
	if (s_search != NULL)
	{
		s_search->interrupt();
	}
	signal(SIGINT, handleSigInt); 
}

bool parseExpression(const char * & c, long * varBase, long & value);

bool parseMember(const char * & c, long * varBase, long & value)
{
	// (<expr>)
	if ('(' == *c)
	{
		++c;
		if (!parseExpression(c, varBase, value))
		{
			return false;
		}

		if (')' != *c)
		{
			return false;
		}
		++c;
		return true;
	}	

	// -<member>
	if ('-' == *c)
	{
		++c;
		if (!parseMember(c, varBase, value))
		{
			return false;
		}

		value = -value;
		return true;
	}

	// number
	if (('0' <= *c) && ('9' >= *c))
	{
		value = 0;
		while (('0' <= *c) && ('9' >= *c))
		{
			value = value * 10 + *c - '0';
			++c;
		}
		return true;
	}

	// variable
	if ((('A' <= *c) && ('Z' >= *c))
		|| (('a' <= *c) && ('z' >= *c)))
	{
		if (NULL == varBase)
		{
			// variables not supported here
			return false;
		}

		long n;
		if (('a' <= *c) && ('z' >= *c))
		{
			n = *c - 'a';
		}
		else
		{
			n = *c - 'A';
		}
		++c;
		value = varBase[n];
		return true;
	}

	return false;
}

bool parseTerm(const char * & c, long * varBase, long & value)
{
	if (!parseMember(c, varBase, value))
	{
		return false;
	}

	while (('*' == *c) || ('/' == *c) || ('%' == *c) || ('@' == *c))
	{
		char op = *c;
		++c;

		long m;

		if (!parseMember(c, varBase, m))
		{
			return false;
		}

		if ('*' == op)
		{
			value = value * m;
		}
		else if (0 == m)
		{
			Writer::print("Division by zero\n");
			return false;
		}
		else if ('/' == op)
		{
			value = value / m;
		}
		else if ('%' == op)
		{
			value = value % m;
		}
		else if ('@' == op)
		{
			value = value % m;
			while (((m > 0) && (value < 0))
				|| ((m < 0) && (value > 0)))
			{
				value += m;
			}
		}
	}

	return true;
}


bool parseExpression(const char * & c, long * varBase, long & value)
{
	if (!parseTerm(c, varBase, value))
	{
		return false;
	}

	while (('+' == *c) || ('-' == *c))
	{
		char op = *c;
		++c;

		long t;

		if (!parseTerm(c, varBase, t))
		{
			return false;
		}

		if ('+' == op)
		{
			value = value + t;
		}
		else
		{
			value = value - t;
		}
	}
	return true;
}

bool parseTarget(const char * & c, long * varBase, Direction & dir, long & dt, long & dx, long & dy)
{
	//<dir><expr>[<expr>,<expr]

	switch (*c)
	{
	case 'F':
		++c;
		if ('x' == *c)
		{
			++c;
			dir = TFX;
		}
		else
		{
			dir = TF;
		}
		break;

	case 'L':
		++c;
		if ('x' == *c)
		{
			++c;
			dir = TLX;
		}
		else
		{
			dir = TL;
		}
		break;

	case 'R':
		++c;
		if ('x' == *c)
		{
			++c;
			dir = TRX;
		}
		else
		{
			dir = TR;
		}
		break;

	case 'B':
		++c;
		if ('x' == *c)
		{
			++c;
			dir = TBX;
		}
		else
		{
			dir = TB;
		}
		break;

	default: 
		return false;
	}
	
	if (!parseExpression(c, varBase, dt))
	{
		return false;
	}

	if ('[' != *c)
	{
		return false;
	}

	++c;

	if (!parseExpression(c, varBase, dx))
	{
		return false;
	}

	if (',' != *c)
	{
		return false;
	}

	++c;

	if (!parseExpression(c, varBase, dy))
	{
		return false;
	}

	if (']' != *c)
	{
		return false;
	}

	++c;

	return true;
}

bool parseVariable(const char * & c, long * varBase, long & var, long & minValue, long & maxValue)
{
	// <var>=<expr>{..<expr>}
	if ((*c >= 'A') && (*c <= 'Z'))
	{
		var = *c - 'A';
	}
	else if ((*c >= 'a') && (*c <= 'z'))
	{
		var = *c - 'a';
	}
	else
	{
		return false;
	}
	++c;
	if ('=' != *c)
	{
		return false;
	}
	++c;
	if (!parseExpression(c, varBase, minValue))
	{
		return false;
	}
	if ('.' != *c)
	{
		// looks like single-expression definition
		maxValue = minValue;
		return true;
	}
	++c;
	if ('.' != *c)
	{
		return false;
	}
	++c;
	if (!parseExpression(c, varBase, maxValue))
	{
		return false;
	}
	if (minValue > maxValue)
	{
		long l = minValue;
		minValue = maxValue;
		maxValue = l;
	}
	return true;
}

bool parseExtendedTarget(Search * srch, const char * c, long sp, long sc, long su, long sl, long sx, long sy, long st)
{
	// extended switch syntax:
	// {(<variable>=<expr>..<expr>,...)}{<dir><expr:time>[<expr:X>,<expr:Y>]-}<dir><expr:time>[<expr:X>,<expr:Y>]
	// example:
	// (A=1..10)F0[10,10+A]-Fx150[50+4*A,50-4*A]

	const char * base = c;

	enum
	{
		numVars = 'Z' - 'A' + 1
	};

	long var[numVars];
	long minVar[numVars];
	long maxVar[numVars];

	const char * parserBase;

	{
		for (long i = 0; i < numVars; i++)
		{
			var[i] = 0;
			minVar[i] = 0;
			maxVar[i] = 0;
		}
	}

	{
		long i;
		i = 'P' - 'A';
		minVar[i] = maxVar[i] = var[i] = sp;

		i = 'C' - 'A';
		minVar[i] = maxVar[i] = var[i] = sc;

		i = 'U' - 'A';
		minVar[i] = maxVar[i] = var[i] = su;

		i = 'L' - 'A';
		minVar[i] = maxVar[i] = var[i] = sl;

		i = 'X' - 'A';
		minVar[i] = maxVar[i] = var[i] = sx;

		i = 'Y' - 'A';
		minVar[i] = maxVar[i] = var[i] = sy;

		i = 'T' - 'A';
		minVar[i] = maxVar[i] = var[i] = st;
	}

	if ('(' == *c)
	{
		do
		{
			++c; // skip '('  or ','

			// parse variable definition
			long varNum, minValue, maxValue;
			if (!parseVariable(c, var, varNum, minValue, maxValue))
			{
				goto error;
			}

			minVar[varNum] = minValue;
			maxVar[varNum] = maxValue;
			var[varNum] = minValue;

		} while (',' == *c); // repeat until last variable is parsed

		if (')' != *c)
		{
			goto error;
		}
		++c; // skip ')'
	}

	parserBase = c; // store target definition address

	{
		long i;
		do
		{
			// parse the target body for each variable combination
			c = parserBase;
			Direction dir1, dir2;
			long dt1, dt2, dx1, dx2, dy1, dy2;
			
			if (!parseTarget(c, var, dir1, dt1, dx1, dy1))
			{
				goto error;
			}

			if ('.' != *c)
			{
				// single-point target
				dir2 = dir1;
				dx2 = dx1;
				dy2 = dy1;
				dt2 = dt1;

				dir1 = TF;
				dx1 = 0;
				dy1 = 0;
				dt1 = 0;

				if ('\0' != *c)
				{
					fprintf(stderr, "'..' expected\n");
				}
			}
			else
			{
				++c;
				if ('.' != *c)
				{
					fprintf(stderr, "'..' expected\n");
				}
				++c;
				if (!parseTarget(c, var, dir2, dt2, dx2, dy2))
				{
					goto error;
				}
			}

			if ('\0' != *c)
			{
				fprintf(stderr, "Extra characters at the end of the line\n");
				goto error;
			}


			char buffer[1024];
			buffer[0] = 0;

			for (i = 0; i < numVars; i++)
			{
				if (var[i] || minVar[i] != maxVar[i])
				{
					if (!buffer[0]) strcpy(buffer, " (");
					else strcat(buffer, ",");

					sprintf(&buffer[strlen(buffer)], "%c=%d", 'a' + i, var[i]);
				}
			}

			if (buffer[0]) strcat(buffer, ")");

			char *str = new char[strlen(buffer) + 1];
			strcpy(str, buffer);

			// add the calculated target

			srch->addTarget(dx1, dy1, dt1, dir1, dx2, dy2, dt2, dir2, str);

			// calculate next variable combination
			i = 0;

			do
			{
				if (var[i] < maxVar[i])
				{
					++var[i];
					break;
				}
				else
				{
					var[i] = minVar[i];
					++i;
				}
			} while (i < numVars);

		} while (i < numVars);
	}

	// all targets added

	return true;

error:

	fprintf(stderr, "Extended target parsing error\n");

	fputs(base, stderr);
	fputc('\n', stderr);

	while (base != c)
	{
		fputc('.', stderr);
		++base;
	}
	fputc('^', stderr);
	fputc('\n', stderr);

	return false;
}

int main(int argc, char * argv[])
{
	long dx = 0;
	long dy = 0;
	long dt = 0;
	Direction dir = TF;
	bool sorted = false;

	long period = 0;
	long compression = 0;
	long minlen = 1;
	long maxlen = 1500;
	long maxcells = 0;
	long maxresults = 0;
	long ignore = 0;
	bool useMemory = false;

	long treeSize = 1000;

	bool help = false;
	bool rleSet = false;
	bool script = false;

	// Create the search object
	s_search = new Search();

	for (int argi = 1; argi < argc; argi++)
	{
		char * ep;

		if (0 == strcmp("-p", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -p switch\n");
				help = true;
			}
			else
			{
				period = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -p switch\n");
					help = true;
				}
			}
		}
		else if (0 == strcmp("-c", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -c switch\n");
				help = true;
			}
			else
			{
				compression = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -c switch\n");
					help = true;
				}
			}
		}
		else if (0 == strcmp("-n", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -n switch\n");
				help = true;
			}
			else
			{
				maxresults = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -n switch\n");
					help = true;
				}
			}
		}
		else if (0 == strcmp("-l", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -l switch\n");
				help = true;
			}
			else
			{
				minlen = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -l switch\n");
					help = true;
				}
				if (minlen > maxlen)
				{
					maxlen = minlen;
				}
				if (minlen < 1)
				{
					minlen = 1;
				}
			}
		}
		else if (0 == strcmp("-i", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -i switch\n");
				help = true;
			}
			else
			{
				ignore = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -i switch\n");
					help = true;
				}
			}
		}
		else if (0 == strcmp("-u", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -u switch\n");
				help = true;
			}
			else
			{
				maxlen = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -u switch\n");
					help = true;
				}
				if (maxlen < minlen)
				{
					minlen = maxlen;
				}
			}
		}
		else if (0 == strcmp("-t", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -t switch\n");
				help = true;
			}
			else
			{
				treeSize = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -t switch\n");
					help = true;
				}
				if (treeSize < 0)
				{
					treeSize = 0;
				}
				else if (treeSize > 100000)
				{
					treeSize = 100000;
				}
			}
		}
		else if (0 == strcmp("-x", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -x switch\n");
				help = true;
			}
			else
			{
				maxcells = strtoul(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid argument for -x switch\n");
					help = true;
				}
			}
		}
		else if (0 == strcmp("-f", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -f switch\n");
				help = true;
			}
			else
			{
				Writer::setFilename(argv[argi]);
			}
		}
		else if (0 == strcmp("-o", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -o switch\n");
				help = true;
			}
			else
			{
				Writer::setRleFilename(argv[argi]);
				rleSet = true;
			}
		}
		else if (0 == strcmp("-s", argv[argi]))
		{
			sorted = true;
		}
		else if (0 == strcmp("-m", argv[argi]))
		{
			useMemory = true;
		}
		else if ((0 == strcmp("-h", argv[argi])) || (0 == strcmp("-?", argv[argi])))
		{
			help = true;
		}
		else if (0 == strcmp("-d", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing DX argument for -d switch\n");
				help = true;
			}
			else
			{
				dx = strtol(argv[argi], &ep, 10);
				if (*ep != '\0')
				{
					fprintf(stderr, "Invalid DX argument for -d switch\n");
					help = true;
				}

				if (++argi >= argc)
				{
					fprintf(stderr, "Missing DY argument for -d switch\n");
					help = true;
				}
				else
				{
					dy = strtol(argv[argi], &ep, 10);
					if (*ep != '\0')
					{
						fprintf(stderr, "Invalid DY argument for -d switch\n");
						help = true;
					}

					if (++argi >= argc)
					{
						fprintf(stderr, "Missing DT argument for -d switch\n");
						help = true;
					}
					else
					{
						dt = strtol(argv[argi], &ep, 10);
						if (*ep != '\0')
						{
							fprintf(stderr, "Invalid DT argument for -d switch\n");
							help = true;
						}

						if (++argi >= argc)
						{
							fprintf(stderr, "Missing DIR argument for -d switch\n");
							help = true;
						}
						else
						{
							if (0 == strcmp("F", argv[argi]))
							{
								dir = TF;
							}
							else if (0 == strcmp("Fx", argv[argi]))
							{
								dir = TFX;
							}
							else if (0 == strcmp("L", argv[argi]))
							{
								dir = TL;
							}
							else if (0 == strcmp("Lx", argv[argi]))
							{
								dir = TLX;
							}
							else if (0 == strcmp("R", argv[argi]))
							{
								dir = TR;
							}
							else if (0 == strcmp("Rx", argv[argi]))
							{
								dir = TRX;
							}
							else if (0 == strcmp("B", argv[argi]))
							{
								dir = TB;
							}
							else if (0 == strcmp("Bx", argv[argi]))
							{
								dir = TBX;
							}
							else
							{
								fprintf(stderr, "Invalid DIR argument for -d switch\n");
								help = true;
							}
						}
					}
				}
			}
		}
		else if (0 == strcmp("-e", argv[argi]))
		{
			if (++argi >= argc)
			{
				fprintf(stderr, "Missing argument for -e switch\n");
				help = true;
			}
			else
			{
				script = true;
				if (!parseExtendedTarget(s_search, argv[argi], period, compression, maxlen, minlen, dx, dy, dt))
				{
					fprintf(stderr, "Invalid argument for -e switch\n");
					help = true;
				}
			}
		}
		else
		{
			fprintf(stderr, "Invalid parameter: %s\n", argv[argi]);
			help = true;
		}
	}

	if ((period == 0) && !rleSet)
	{
		help = true;
	}

	if (compression == 0)
	{
		compression = period;
	}

	if (help)
	{
		fprintf(stderr, 
			"Usage:\n"
			"    hersrch [-p period] [-c compression] [-d dx dy dt dir] [-e <script>] [-t treesizeKB] [-m] [-l minlen] [-u maxlen] [-x maxcell] [-n maxresults] [-f outfile] [-o rlefile] [-s] [-i ignoresets]\n");

		return 0;
	}

	if (period == 0)
	{
		// generate test pattern
		s_search->test();
	}
	else
	{
		if (!script)
		{
			// generate implicit targets
			for (long i = dt; i <= maxlen; i += period)
			{
				if (i >= minlen)
				{
					s_search->addTarget(dx, dy, i, dir);
				}
			}
		}
		s_search->trimTargets(maxlen);

		// Install signal handler

		signal(SIGINT, handleSigInt); 

		// do the search
		s_search->run(
			period, compression, 
			maxcells, maxresults, 
			sorted, treeSize, ignore, useMemory);
	}

	return 0;
}

